-- Set up variables used globally by the addon
local IconFrameContainer -- Icon container
local TextboxFrame -- Text container.
local Buffs --  A table containing the RTB buffs and their state, name, short name and their corresponding frames.
local BuffTextures -- To retrieve the textures for each buff.
local OnUpdateFrame = CreateFrame("frame")
local EventFrame -- The main event frame to catch all registered events.
local Initialized = false -- Addon initialisation state to prevent multiple initializes.
local InitialIconSize = 40 -- Standard icon size when scaling factor is 1.
local PlayerIsInCombat = false -- Shows if the player is in a fight.
local isRogue = false
local isOutlaw = false
local configMode = false -- Is set during GUI configuration with /ttb. Prevents the frames from hiding.

--[[ Indicates for the TTB_OnUpdate function that all cooldown animations have to be started anew. Is set by the
     UNIT_SPELLCAST_SUCCEEDED event handler. Setting this to true per default enables us
	 to directly draw a cooldown icon without using RTB if we /reload and buffs are already running.
--]]
local JustRolled = true

-- RTB spell IDs.
local RollTheBones = 315508
local GrandMelee = "193358"
local RuthlessPrecision = "193357"
local TrueBearing = "193359"
local Broadsides = "193356"
local SkullAndCrossbones = "199603"
local BuriedTreasure = "199600"

--[[ Makes the given frame movable.

	 @param Frame
--]]
function TTB_MakeFrameMovable(Frame)
	Frame:SetMovable(true)
	Frame:RegisterForDrag("LeftButton")
	Frame:SetScript("OnDragStart", Frame.StartMoving)
end

--[[ Creates an event handler for the given frame.
	 It lets a frame snap to the edges of the screen when it (partly) overlaps them.

	 @param Frame
--]]
function TTB_MakeFrameSnappable(Frame)
	Frame:SetScript("OnDragStop", function(self, event, ...)
		Frame:StopMovingOrSizing()
		local x = Frame:GetLeft()
		local y = Frame:GetBottom()

		if(x < 0) then
			x = 0
		elseif(x + Frame:GetWidth() > GetScreenWidth()) then
			x = GetScreenWidth() - Frame:GetWidth()
		end

		if(y < 0) then
			y = 0
		elseif(y + Frame:GetHeight() > GetScreenHeight()) then
			y = GetScreenHeight() - Frame:GetHeight()
		end

		Frame:ClearAllPoints()
		Frame:SetPoint("BOTTOMLEFT", x, y)
	end)
end

--[[ The icon container. This one holds the graphical buff representations.
     We cannot initialise the text/icon frames directly after PLAYER_LOGIN because
	 their saved positioning will then get lost.
--]]
IconFrameContainer = CreateFrame("Frame", "TTB_IconFrameContainer", UIParent, "BackdropTemplate")
IconFrameContainer:SetPoint("CENTER")
IconFrameContainer:SetFrameStrata("LOW")
IconFrameContainer:Hide()

local CooldownBarFrame = CreateFrame("Frame", "CooldownBarFrame", IconFrameContainer)
local CooldownBarFrameLine = CooldownBarFrame:CreateLine()

TTB_MakeFrameMovable(IconFrameContainer)
TTB_MakeFrameSnappable(IconFrameContainer)

--[[ The text container. Using a MessageFrame resulted in a flickering display
	 after switching to OnUpdating its content every 100ms.
--]]
TextboxFrame = CreateFrame("Frame", "TTB_TextboxFrame", UIParent, "BackdropTemplate")
TextboxFrame:SetPoint("CENTER")
TextboxFrame:SetSize(118, 135)

TextboxFrame:SetBackdrop({
	bgFile = [[Interface\DialogFrame\UI-DialogBox-Background]],
	tile = true, tileSize = 16, edgeSize = 16,
	insets = { left = 1, right = 1, top = 1, bottom = 1 }
})

TTB_MakeFrameMovable(TextboxFrame)
TTB_MakeFrameSnappable(TextboxFrame)
TextboxFrame:Hide()

--[[ This function creates an icon which is put into the IconFrameContainer.
     The icon is positioned and sized via TTB_DrawIcons() and refreshed
	 by TTB_OnUpdate().

		@param FrameName			Name of the frame (rather cosmetic)
		@param TextureFile			Link to the texture file to use for the icon

		@return WrapperFrame		Wrapper frame for the cooldown animation
		@return CooldownFrame		Cooldown animation frame to be placed on the WrapperFrame
--]]
function TTB_CreateIconFrames(FrameName, TextureFile)
	local WrapperFrame = CreateFrame("Frame", FrameName, IconFrameContainer)
	WrapperFrame:SetAlpha(.25)

	WrapperFrame.texture = WrapperFrame:CreateTexture()
	WrapperFrame.texture:SetAllPoints()
	WrapperFrame.texture:SetTexture(TextureFile)

	local CooldownFrame = CreateFrame("Cooldown", FrameName .. "Cooldown", WrapperFrame, "CooldownFrameTemplate")
	CooldownFrame:SetReverse(true)
	CooldownFrame:SetHideCountdownNumbers(true)

	CooldownFrame.text = CooldownFrame:CreateFontString(nil, "OVERLAY", "GameFontNormalLarge")
	CooldownFrame.text:SetPoint("CENTER")
	CooldownFrame.text:SetText("")
	CooldownFrame.text:SetTextColor(1, 1, 1, 1)

	return WrapperFrame, CooldownFrame
end

--[[ This function creates the frames for the text mode which are in the TextboxFrame.
	 The frames are positioned via TTB_DrawTextbox() and refreshed by TTB_OnUpdate().

		@param FrameName			Name of the frame (rather cosmetic)
		@param TextureFile			Link to the texture file to use for the icon

		@return IconFrame			The icon displayed left to the text
		@return TextFrame			The text right to the icon showing the short buff description
--]]
function TTB_CreateTextboxFrames(FrameName, TextureFile)
	local IconFrame = CreateFrame("Frame", FrameName, TextboxFrame)

	IconFrame.texture = IconFrame:CreateTexture()
	IconFrame.texture:SetAllPoints()
	IconFrame.texture:SetTexture(TextureFile)

	local TextFrame = CreateFrame("Frame", FrameName .. "Text", TextboxFrame)

	TextFrame.text = TextFrame:CreateFontString(nil, "OVERLAY")
	TextFrame.text:SetFont("Fonts\\FRIZQT__.TTF", 13, "OUTLINE")
	TextFrame.text:SetPoint("LEFT", 0, 0)
	TextFrame.text:SetTextColor(1, 1, 1, 1)
	TextFrame.text:SetText("")

	IconFrame:SetAlpha(.25)
	TextFrame:SetAlpha(.25)
	
	return IconFrame, TextFrame
end

-- This function updates the icon and text box presentation regularly.
local updateInterval = 0.1 -- Update the frames every 0.1 seconds. A bigger number sometimes results in a noticeable delay after rolling.
local runningBuffs = {} -- Holds the buffs currently running.
local lastRunningBuffs = {} -- Holds the buffs that were active the last time TTB_OnUpdate() ran.

function TTB_OnUpdate(self, elapsed)
	self.TimeSinceLastUpdate = (self.TimeSinceLastUpdate or 0) + elapsed;

	if(self.TimeSinceLastUpdate < updateInterval) then
		return
	end

	self.TimeSinceLastUpdate = 0;

	--[[ Reset buffs. It's easier to pretend no buff is up before detection. So we only have
	     to set the ones to active which are up. The other ones can stay inactive.
	--]]
	for key, value in ipairs(TTBDB.BuffsOrdered) do
		Buffs[value]["active"] = false
	end

	runningBuffs = {}

	--[[ Check if buffs are active and update buff table.
		 After Patch 8.0.1 we cannot use UnitBuff with a spell name any more, so we have to check every player buff.
	--]]
	local AtLeastOneRTBBuffActive = false
	
	for i = 1, 40 do
		local name, icon, count, dispelType, duration, expires, caster, isStealable, nameplateShowPersonal, spellId, canApplyAura, isBossDebuff, _, nameplateShowAll, timeMod, value1, value2, value3 = UnitBuff("player" , i)
		spellId = tostring(spellId)

		if(Buffs[spellId] ~= nil) then
			Buffs[spellId]["active"] = true

			--[[ To correctly set a cooldown animation for RTB buffs, which come "out of nowhere"
				 (e. g. from T21 4p), we check if JustRolled is false (so we make sure those new buff(s)
				 don't just come from RTB) and compare the previously running buffs (from the last TTB_OnUpdate call)
				 with the current ones. Every new buff has to be a sneaky little bastard.
			--]]
			if(not JustRolled and (lastRunningBuffs[spellId] == nil or lastRunningBuffs[spellId] == false)) then
				Buffs[spellId]["iconCdFrame"]:SetCooldown(GetTime(), math.ceil(expires - GetTime()))
			end

			runningBuffs[spellId] = true

			AtLeastOneRTBBuffActive = true
		end
	end

	lastRunningBuffs = runningBuffs

	--[[ Update the icon bar and set cooldowns. Handle the display of the icon bar
	     according to the settings.
	--]]
	if(TTBDB.iconmode) then
		for key, value in ipairs(TTBDB.BuffsOrdered) do
			if(Buffs[value]["active"]) then
				for i = 1, 40 do
					local name, icon, count, dispelType, duration, expires, caster, isStealable, nameplateShowPersonal, spellId, canApplyAura, isBossDebuff, _, nameplateShowAll, timeMod, value1, value2, value3 = UnitBuff("player" , i)
					spellId = tostring(spellId)

					if(spellId == value) then
						Buffs[value]["iconFrame"]:SetAlpha(1)

						-- Only update the cooldown animation if a roll was done beforehand, else it would start on every frame.
						if(JustRolled) then
							Buffs[value]["iconCdFrame"]:SetCooldown(GetTime(), math.ceil(expires - GetTime()))
						end

						Buffs[value]["iconCdFrame"].text:SetText(math.ceil(expires - GetTime()))
						Buffs[value]["iconCdFrame"].text:SetAlpha(1)

						break
					end
				end
			else
				Buffs[value]["active"] = false
				Buffs[value]["iconFrame"]:SetAlpha(.25)
				Buffs[value]["iconCdFrame"]:SetCooldown(0, 0)
				Buffs[value]["iconCdFrame"].text:SetText("")
				Buffs[value]["iconCdFrame"].text:SetAlpha(0)
			end

		-- Display of the RTB cooldown, when icon background is enabled

		local RTBOnCooldown = false

		if(TTBDB.iconbackground == true) then
			if(TTBDB.icondir == "h") then
				local IconBoxWidth = IconFrameContainer:GetSize()

				CooldownBarFrame:SetPoint("BOTTOMLEFT", IconFrameContainer, "BOTTOMLEFT", 0, -2.5)
				CooldownBarFrame:SetSize(IconBoxWidth, 10)
				CooldownBarFrameLine:SetThickness(5)

				local start, duration = GetSpellCooldown(RollTheBones)
				local NewIconBoxWidth = IconBoxWidth 
				if(start > 0 and duration > 0 and duration == 45) then
					local secondsLeftOfCooldown = start + duration - GetTime()
					local percentageOfCooldown = 1 - (secondsLeftOfCooldown / duration)
					RTBOnCooldown = true

					NewIconBoxWidth = IconBoxWidth * percentageOfCooldown
					CooldownBarFrameLine:SetColorTexture(241/255, 128/255, 45/255, 1)
				elseif(start == 0) then
					CooldownBarFrameLine:SetColorTexture(13/255, 69/255, 50/255, 1)
				end

				CooldownBarFrameLine:SetStartPoint("BOTTOMLEFT", 0, 0)
				CooldownBarFrameLine:SetEndPoint("BOTTOMLEFT", NewIconBoxWidth, 0)
			elseif(TTBDB.icondir == "v") then
				local _, IconBoxHeight = IconFrameContainer:GetSize()

				CooldownBarFrame:SetPoint("BOTTOMLEFT", IconFrameContainer, "BOTTOMLEFT", -2.5, 0)
				CooldownBarFrame:SetSize(10, IconBoxHeight)
				CooldownBarFrameLine:SetThickness(5)

				local start, duration = GetSpellCooldown(RollTheBones)
				local NewIconBoxHeight = IconBoxHeight 
				if(start > 0 and duration > 0 and duration == 45) then
					local secondsLeftOfCooldown = start + duration - GetTime()
					local percentageOfCooldown = 1 - (secondsLeftOfCooldown / duration)
					RTBOnCooldown = true

					NewIconBoxHeight = IconBoxHeight * percentageOfCooldown
					CooldownBarFrameLine:SetColorTexture(241/255, 128/255, 45/255, 1)
				elseif(start == 0) then
					CooldownBarFrameLine:SetColorTexture(13/255, 69/255, 50/255, 1)
				end

				CooldownBarFrameLine:SetStartPoint("BOTTOMLEFT", 0, 0)
				CooldownBarFrameLine:SetEndPoint("BOTTOMLEFT", 0, NewIconBoxHeight)
			end
		end
		
		-- Show the icon bar if we have at least one active buff, we are in combat, the icons are pinned, we are in configMode or RTB is on cooldown.
		if(AtLeastOneRTBBuffActive or PlayerIsInCombat or TTBDB.pinicons or configMode or RTBOnCooldown) then
			IconFrameContainer:Show()
		else
			IconFrameContainer:Hide()
		end
	end

	elseif(configMode) then
		IconFrameContainer:Show()
	end

	if(TTBDB.textboxmode) then
		for key, value in ipairs(TTBDB.BuffsOrdered) do
			local IconFrame = Buffs[value]["tbIconFrame"]
			local TextFrame = Buffs[value]["tbTextFrame"]

			if(Buffs[value]["active"]) then
				IconFrame:SetAlpha(1)
				TextFrame:SetAlpha(1)
				TextFrame.text:SetTextColor(0.34, 0.68, 0.41, 1)
			else
				TextFrame.text:SetTextColor(1, 1, 1, 1)
				IconFrame:SetAlpha(.25)
				TextFrame:SetAlpha(.25)
			end
		end

		-- Show the text box if we have at least one active buff, we are in combat, the icons are pinned or we are in configMode.
		if(AtLeastOneRTBBuffActive or PlayerIsInCombat or TTBDB.pintextbox or configMode) then
			TextboxFrame:Show()
		else
			TextboxFrame:Hide()
		end

	elseif(configMode) then
		TextboxFrame:Show()
	end

	JustRolled = false
end

-- Main Event listener.
EventFrame = CreateFrame("Frame")
EventFrame:RegisterEvent("ADDON_LOADED")

EventFrame:SetScript("OnEvent", function(self, event, ...)
	if(event == "ADDON_LOADED" and ... == "TrollTheBones") then
		-- Initialize saved variables if called for the first time.
		TTB_SetStandardValues()

		--[[ We have to use PLAYER_LOGIN to be able to access some data like GetSpecialization().
			 ADDON_LOADED kicks in too early during the game loading process.
		--]]
		EventFrame:RegisterEvent("PLAYER_LOGIN")

	elseif(event == "PLAYER_LOGIN") then
		if(Initialized == false) then
			TTB_init()
		end

	-- If we respec, check if spec is outlaw and enable/disable the addon.
	elseif(event == "ACTIVE_TALENT_GROUP_CHANGED") then
		if(GetSpecialization() == 2) then
			isOutlaw = true

			EventFrame:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
			EventFrame:RegisterEvent("PLAYER_REGEN_ENABLED")
			EventFrame:RegisterEvent("PLAYER_REGEN_DISABLED")

			OnUpdateFrame:SetScript("OnUpdate", TTB_OnUpdate)

			if(TTBDB.pinicons) then
				IconFrameContainer:Show()
			end

			if(TTBDB.pintextbox) then
				TextboxFrame:Show()
			end
		else
			isOutlaw = false

			EventFrame:UnregisterEvent("UNIT_SPELLCAST_SUCCEEDED")
			EventFrame:UnregisterEvent("PLAYER_REGEN_ENABLED")
			EventFrame:UnregisterEvent("PLAYER_REGEN_DISABLED")

			OnUpdateFrame:SetScript("OnUpdate", nil)

			if(IconFrameContainer ~= nil) then
				IconFrameContainer:Hide()
			end

			if(TextboxFrame ~= nil) then
				TextboxFrame:Hide()
			end
		end

	-- Set the PlayerIsInCombat status.
	elseif(event == "PLAYER_REGEN_ENABLED") then
		PlayerIsInCombat = false

	elseif(event == "PLAYER_REGEN_DISABLED") then
		PlayerIsInCombat = true

	--[[ Main RTB buff detection. This is called after every RTB cast.
	     We check, which buffs are up and accordingly play sounds and
		 update the roll statistics. UI changes are solely done in the TTB_OnUpdate function.
	--]]
	elseif(event == "UNIT_SPELLCAST_SUCCEEDED") then
		local unitId, spellString, spellId = ...

		if(unitId ~= "player" or spellId ~= RollTheBones) then -- Check if player rolled the bones: return if not.
			return
		end

		JustRolled = true
		local amount = 0 -- Number of buffs active.
		TTBDB.rolls = TTBDB.rolls + 1
		TTBDB.lastfive = TTBDB.lastfive + 1

		--[[ Check how many RTB buffs are active. Walk through all player buffs.
			 After Patch 8.0.1 we cannot use UnitBuff with a spell name any more.
			 
			 We only want to count the buffs here that come from Roll The Bones usage. Else any
			 extra buffs like those of the T21 4p set bonus will result in a wrong number of buffs calculated.
			 As of now, we regard any buff with a duration of <= 10 seconds not as an RTB buff, since using RTB always
			 creates buffs at least 12 seconds long.
		--]]
		for i = 1, 40 do
			local name, icon, count, dispelType, duration, expires, caster, isStealable, nameplateShowPersonal, spellId, canApplyAura, isBossDebuff, _, nameplateShowAll, timeMod, value1, value2, value3 = UnitBuff("player" , i)
			spellId = tostring(spellId)

			if(Buffs[spellId] ~= nil and duration > 15) then
				amount = amount + 1
			end
		end

		if(amount > 1) then -- Actions for two or more buffs.
			-- Message after ending a bad luck streak.
			if(TTBDB.rolls >= TTBDB.threshold) then
				if(TTBDB.outputchannel ~= "none") then
					-- 8.2.5 does not allow saying or yelling outside of instanced content any more
					if(IsInInstance() == true or (TTBDB.outputchannel ~= "say" and TTBDB.outputchannel ~= "yell")) then
						SendChatMessage(string.gsub(TTBDB.finallyMessage, '{n}', TTBDB.rolls), TTBDB.outputchannel)
					end
				end

				-- Update record.
				if(TTBDB.rolls > TTBDB.record) then
					TTBDB.record = TTBDB.rolls
				end
			end

			TTBDB.rolls = 0
		end

		 -- Actions for five buffs.
		if(amount == 5) then
			-- Update lastfiverecord.
			if(TTBDB.lastfive > TTBDB.lastfiverecord) then
				TTBDB.lastfiverecord = TTBDB.lastfive
			end

			if(TTBDB.outputchannel ~= "none") then
				-- 8.2.5 does not allow saying or yelling outside of instanced content any more
				if(IsInInstance() == true or (TTBDB.outputchannel ~= "say" and TTBDB.outputchannel ~= "yell")) then
					SendChatMessage(string.gsub(TTBDB.hamMessage, '{n}', TTBDB.lastfive), TTBDB.outputchannel)
				end
			end

			-- Play Leeroy.
			if(TTBDB.leeroy) then
				PlaySoundFile("Interface\\AddOns\\TrollTheBones\\sounds\\leeroy-jenkins.ogg", "Master")
			end

			local LastFiveDateText = ""
			if(TTBDB.lastfivedate > 0) then
				LastFiveDateText = " Your last five roll was on " .. date("%m/%d/%y %H:%M:%S", TTBDB.lastfivedate) .. "."
			end

			-- Print to chat.
			print("You've rolled 'em all!" .. LastFiveDateText .. " You needed " .. TTBDB.lastfive .. " rolls for this.")

			TTBDB.rolls = 0
			TTBDB.lastfive = 0
			TTBDB.lastfivedate = time()
		end

		-- Play sounds.
		if(TTBDB.playsound) then
			if(amount == 1) then
				PlaySoundFile("Interface\\AddOns\\TrollTheBones\\sounds\\1-one.ogg", "Master")
			elseif(amount == 2) then
				PlaySoundFile("Interface\\AddOns\\TrollTheBones\\sounds\\2-two.ogg", "Master")
			elseif(amount == 5) then
				if(TTBDB.leeroy == false) then -- Only play if leeroy mode is set to false
					PlaySoundFile("Interface\\AddOns\\TrollTheBones\\sounds\\5-five.ogg", "Master")
				end
			end
		end
	end
end)

--[[ This function initializes the whole addon.
     It sets all EventHandlers and variables and creates all frames.
--]]
function TTB_init()
	Initialized = true

	local classDisplayName, class, classId = UnitClass("player")
	if(classId == 4) then
		isRogue = true

		if(GetSpecialization() == 2) then
			isOutlaw = true
		end
	end

	-- Lock or unlock the frames depending on configuration.
	if(TTBDB.lockicons) then
		IconFrameContainer:EnableMouse(false)
	else
		IconFrameContainer:EnableMouse(true)
	end

	if(TTBDB.locktextbox) then
		TextboxFrame:EnableMouse(false)
	else
		TextboxFrame:EnableMouse(true)
	end

	-- Show or hide the frames depending on configuration and only if on Outlaw.
	if(TTBDB.pinicons and isOutlaw) then
		IconFrameContainer:Show()
	end

	if(TTBDB.pintextbox and isOutlaw) then
		TextboxFrame:Show()
	end

	-- For retrieving the buff's textures
	BuffTextures = {
		[GrandMelee] = "Interface\\Icons\\Ability_rogue_rollthebones02",
		[RuthlessPrecision] = "Interface\\Icons\\Ability_rogue_rollthebones03",
		[TrueBearing] = "Interface\\Icons\\Ability_rogue_rollthebones04",
		[Broadsides] = "Interface\\Icons\\Ability_rogue_rollthebones07",
		[SkullAndCrossbones] = "Interface\\Icons\\Ability_rogue_rollthebones01",
		[BuriedTreasure] = "Interface\\Icons\\Ability_rogue_rollthebones05",
	}

	--[[ Create the frames for the icon mode. --]]
	-- Grand Melee frames.
	local GMWrapperFrame, GMCooldownFrame = TTB_CreateIconFrames("TTB_GMCooldownFrame", BuffTextures[GrandMelee])

	-- Ruthless Precision frames.
	local RPWrapperFrame, RPCooldownFrame = TTB_CreateIconFrames("TTB_RPCooldownFrame", BuffTextures[RuthlessPrecision])

	-- True Bearing frames.
	local TBWrapperFrame, TBCooldownFrame = TTB_CreateIconFrames("TTB_TBCooldownFrame", BuffTextures[TrueBearing])

	-- Broadsides frames.
	local BWrapperFrame, BCooldownFrame = TTB_CreateIconFrames("TTB_BCooldownFrame", BuffTextures[Broadsides])

	-- Skull And Crossbones frames.
	local SACWrapperFrame, SACCooldownFrame = TTB_CreateIconFrames("TTB_SACCooldownFrame", BuffTextures[SkullAndCrossbones])

	-- Buried Treasure frames.
	local BTWrapperFrame, BTCooldownFrame = TTB_CreateIconFrames("TTB_BTCooldownFrame", BuffTextures[BuriedTreasure])

	--[[ Create the frames for the text box mode. --]]
	-- Grand Melee frames.
	local GMTextboxIconFrame, GMTextboxTextFrame = TTB_CreateTextboxFrames("TTB_GMTextboxIconFrame", BuffTextures[GrandMelee])

	-- Ruthless Precision frames.
	local RPTextboxIconFrame, RPTextboxTextFrame = TTB_CreateTextboxFrames("TTB_RPTextboxIconFrame", BuffTextures[RuthlessPrecision])

	-- True Bearing frames.
	local TBTextboxIconFrame, TBTextboxTextFrame = TTB_CreateTextboxFrames("TTB_TBTextboxIconFrame", BuffTextures[TrueBearing])

	-- Broadsides frames.
	local BTextboxIconFrame, BTextboxTextFrame = TTB_CreateTextboxFrames("TTB_BTextboxIconFrame", BuffTextures[Broadsides])

	-- Skull And Crossbones frames.
	local SACTextboxIconFrame, SACTextboxTextFrame = TTB_CreateTextboxFrames("TTB_SACTextboxIconFrame", BuffTextures[SkullAndCrossbones])

	-- Buried Treasure frames.
	local BTTextboxIconFrame, BTTextboxTextFrame = TTB_CreateTextboxFrames("TTB_BTTextboxIconFrame", BuffTextures[BuriedTreasure])


	--[[ This table contains the RTB Buffs and some information attached to them, such as
	     the corresponding frames in the cooldown frame container and the text box.
		 We can later walk through this list and show icons. etc in the order defined by TTBDB.BuffsOrdered.
	--]]
	Buffs = {
		[GrandMelee] = {
			["active"] = false,
			["name"] = "Grand Melee",
			["abbr"] = "SnD + Leech",
			["iconFrame"] = GMWrapperFrame,
			["iconCdFrame"] = GMCooldownFrame,
			["tbIconFrame"] = GMTextboxIconFrame,
			["tbTextFrame"] = GMTextboxTextFrame,
		},
		[RuthlessPrecision] = {
			["active"] = false,
			["name"] = "Ruthless Precision",
			["abbr"] = "Crit",
			["iconFrame"] = RPWrapperFrame,
			["iconCdFrame"] = RPCooldownFrame,
			["tbIconFrame"] = RPTextboxIconFrame,
			["tbTextFrame"] = RPTextboxTextFrame,
		},
		[TrueBearing] = {
			["active"] = false,
			["name"] = "True Bearing",
			["abbr"] = "CD Reduce",
			["iconFrame"] = TBWrapperFrame,
			["iconCdFrame"] = TBCooldownFrame,
			["tbIconFrame"] = TBTextboxIconFrame,
			["tbTextFrame"] = TBTextboxTextFrame,
		},
		[Broadsides] = {
			["active"] = false,
			["name"] = "Broadsides",
			["abbr"] = "+CP",
			["iconFrame"] = BWrapperFrame,
			["iconCdFrame"] = BCooldownFrame,
			["tbIconFrame"] = BTextboxIconFrame,
			["tbTextFrame"] = BTextboxTextFrame,
		},
		[SkullAndCrossbones] = {
			["active"] = false,
			["name"] = "Skull And Crossbones",
			["abbr"] = "SS%",
			["iconFrame"] = SACWrapperFrame,
			["iconCdFrame"] = SACCooldownFrame,
			["tbIconFrame"] = SACTextboxIconFrame,
			["tbTextFrame"] = SACTextboxTextFrame,
		},
		[BuriedTreasure] = {
			["active"] = false,
			["name"] = "Buried Treasure",
			["abbr"] = "Energy Reg",
			["iconFrame"] = BTWrapperFrame,
			["iconCdFrame"] = BTCooldownFrame,
			["tbIconFrame"] = BTTextboxIconFrame,
			["tbTextFrame"] = BTTextboxTextFrame,
		},
	}

	-- Set the text box frame background.
	TTB_SetTextboxBackground(TTBDB.textboxbackground)

	-- Set the icon frame background.
	TTB_SetIconBackground(TTBDB.iconbackground)

	-- Draw and scale the icons.
	TTB_DrawIcons(TTBDB.iconscaling, TTBDB.icondir)

	-- Draw the textbox
	TTB_DrawTextbox()

	-- Init the Buff Order editor in the ACP
	TTB_ACPInitBuffOrder()

	--[[ Set the event handler for a talent switch. We need this to be active on every
		 Rogue spec to be able to react to it.
	--]]
	if(isRogue) then
		EventFrame:RegisterEvent("ACTIVE_TALENT_GROUP_CHANGED")

		-- Only register the event handlers if we are in Outlaw spec
		if(isOutlaw) then
			OnUpdateFrame:SetScript("OnUpdate", TTB_OnUpdate)

			EventFrame:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
			EventFrame:RegisterEvent("PLAYER_REGEN_ENABLED")
			EventFrame:RegisterEvent("PLAYER_REGEN_DISABLED")
		end
	end
end

--[[ Draws the icon box. This depends on a couple of settings. We have to take the
     icon scaling and the orientation of the icon bar into consideration.
	 We pass scaleFactor/orientation instead of using the TTBDB values, because we want
	 to provide the functionality of updating the scaling on the fly for a live preview
	 when using the slider in the options menu without having to change the saved variables.
	 
	 @param scaleFactor [0.5 - 2]
	 @param orientation "h" or "v"
--]]
function TTB_DrawIcons(scaleFactor, orientation)
	-- To prevent calling the function with wrong values internally.
	if(scaleFactor < 0.5 or scaleFactor > 2) then
		print("|cffff6060Internal Error:|r The scale factor must be between 0.5 and 2!")

		return false

	elseif(orientation ~= "h" and orientation ~= "v") then
		print("|cffff6060Internal Error:|r The orientation must only be \"h\" or \"v\"!")

		return false
	end

	-- Size of the new squares.
	local newIconSize = InitialIconSize * scaleFactor
	
	-- Total needed IconFrameContainer width with spaces (.25 of the icons size).
	local newIconFrameContainerWidth = newIconSize * 6 + ((newIconSize * 0.25) * 7)

	-- Total needed IconFrameContainer height with spaces (.25 of the icons size).
	local newIconFrameContainerHeight = newIconSize + ((newIconSize * 0.25) * 2)

	-- Scale the icons and set their positions.
	local xPos = newIconSize * 0.25;
	local yPos = newIconSize * 0.25;

	-- Walk through the ordered Buffs and display them accordingly.
	for key, value in ipairs(TTBDB.BuffsOrdered) do
		local CooldownFrame = Buffs[value]["iconFrame"]
		local Cooldown = Buffs[value]["iconCdFrame"]

		CooldownFrame:SetSize(newIconSize, newIconSize)
		Cooldown:SetSize(newIconSize, newIconSize)

		--[[ Depending on if we want to draw the icons horizontally or vertically,
		     we just always set the other value to a fixed value, so only the x-axis (horizontally)
			 or the y axis (vertically) increases.
		--]]
		if(orientation == "h") then
			CooldownFrame:SetPoint("LEFT", IconFrameContainer, "LEFT", xPos, 0)

			xPos = xPos + newIconSize + (newIconSize * 0.25)
		else
			--[[ This is a bit tricky. Changing the orientation here to "TOP" is buggy (so we have to use "LEFT" just as in horizontal mode).
			     Otherwise the icons would get displayed like a staircase. Because of keeping the "LEFT" orientation,
				 icons start in the vertical middle of the box. We have to move them up by half
				 of the CooldownFrame's height (which is in the width variable) minus the size of an icon.

				 // TODO: This could actually have something to do with rescaling the container afterwards.
			--]]
			CooldownFrame:SetPoint("LEFT", IconFrameContainer, "LEFT", xPos, yPos + (newIconFrameContainerWidth / 2 - newIconSize))

			yPos = yPos - newIconSize - (newIconSize * 0.25)
		end

		Cooldown.text:SetFont("Fonts\\FRIZQT__.TTF", TTB_Round(18 * scaleFactor, 0), "OUTLINE")
	end

	-- Scale the container.
	if(orientation == "h") then
		IconFrameContainer:SetSize(newIconFrameContainerWidth, newIconFrameContainerHeight)
	else
		IconFrameContainer:SetSize(newIconFrameContainerHeight, newIconFrameContainerWidth)
	end
end

--[[ Draws the texbox. Can be used to repaint it after changing e.g. TTBDB.BuffsOrdered.
--]]
function TTB_DrawTextbox()
	local yPos = -10

	for key, value in ipairs(TTBDB.BuffsOrdered) do
		local IconFrame = Buffs[value]["tbIconFrame"]
		local TextFrame = Buffs[value]["tbTextFrame"]

		IconFrame:SetSize(15, 15)
		IconFrame:SetPoint("TOPLEFT", 8, yPos)
		IconFrame.texture:SetTexture(BuffTextures[value])

		TextFrame:SetSize(150, 15)
		TextFrame:SetPoint("TOPLEFT", 28, yPos)
		TextFrame.text:SetText(Buffs[value]["abbr"])

		yPos = yPos - 20;
	end
end

--[[ Shows or hides the background of the CooldownFrame. There is no need
     to do this in TTB_DrawIcons(), so for a better performance we use a function.

     @param showBackground Whether to show the background (boolean)
--]]
function TTB_SetIconBackground(showBackground)
	if(showBackground) then
		IconFrameContainer:SetBackdrop({
				bgFile = [[Interface\DialogFrame\UI-DialogBox-Background]],
				tile = true, tileSize = 16, edgeSize = 16,
				insets = { left = 0, right = 0, top = 0, bottom = 0 }
		})
	else
		IconFrameContainer:SetBackdrop(nil)
	end
end

--[[ Shows or hides the background of the TextboxFrame. There is no need
     to do this in TTB_DrawTextbox(), so for a better performance we use a function.

     @param showBackground Whether to show the background (boolean)
--]]
function TTB_SetTextboxBackground(showBackground)
	if(showBackground) then
		TextboxFrame:SetBackdrop({
				bgFile = [[Interface\DialogFrame\UI-DialogBox-Background]],
				tile = true, tileSize = 16, edgeSize = 16,
				insets = { left = 3, right = 3, top = 5, bottom = 3 }
		})
	else
		TextboxFrame:SetBackdrop(nil)
	end
end

-- Resets the IconFrameContainer.
function TTB_ResetIconFrameContainer()
	for key, value in pairs(Buffs) do
		Buffs[key]["active"] = false
		Buffs[key]["iconFrame"]:SetAlpha(.25)
		Buffs[key]["iconCdFrame"]:SetCooldown(0, 0)
		Buffs[key]["iconCdFrame"].text:SetText("")
		Buffs[key]["iconCdFrame"].text:SetAlpha(0)
	end
end

-- Resets the TextFrame.
function TTB_ResetTextboxFrame()
	for key, value in pairs(Buffs) do
		Buffs[key]["tbTextFrame"].text:SetTextColor(1, 1, 1, 1)
		Buffs[key]["tbIconFrame"]:SetAlpha(.25)
		Buffs[key]["tbTextFrame"]:SetAlpha(.25)
	end
end

-- Addon control panel.
ACPanel = CreateFrame("Frame", "TTB_ACPanel", UIParent)
ACPanel.name = "|TInterface\\Icons\\Ability_rogue_rollthebones:13:13:0:0:128:128:10:118:10:118|t Troll The Bones"
InterfaceOptions_AddCategory(ACPanel)

--[[ This function creates a check button.

	 @param parent 		The parent frame
	 @param x_loc 		X position
	 @param y_loc Y 	position
	 @param displayname Text next to the button

	 @return 			The check button frame
--]]
local i = 1
function TTB_CreateCheckbutton(parent, x_loc, y_loc, displayname)
	i = i + 1
	
	local CheckButton = CreateFrame("CheckButton", "TTB_CheckButton0" .. i, parent, "ChatConfigCheckButtonTemplate")
	CheckButton:SetPoint("TOPLEFT", x_loc, y_loc)
	getglobal(CheckButton:GetName() .. 'Text'):SetText(displayname)

	return CheckButton
end

local LeeroyCheckbox = TTB_CreateCheckbutton(ACPanel, 10, -10, "Leeroy mode")
LeeroyCheckbox.tooltip = "Turn it on and scream along!"

local PlaySoundCheckbox = TTB_CreateCheckbutton(ACPanel, 180, -10, "Play sound")
PlaySoundCheckbox.tooltip = "Check to play a short sound (\"One\" to \"Five\") after gaining bones buffs."

local TextboxModeCheckbox = TTB_CreateCheckbutton(ACPanel, 10, -120, "Text box mode")
TextboxModeCheckbox.tooltip = "When in combat, this option displays a little text box showing the six buffs line by line, roughly ordered from best to worst."

local TextboxBackgroundCheckbox = TTB_CreateCheckbutton(ACPanel, 180, -120, "Text box background")
TextboxBackgroundCheckbox.tooltip = "Show or hide the transparent black background in text box mode."

-- Preview the background of the text box when checking the background box.
TextboxBackgroundCheckbox:SetScript("OnClick", function(self, ...)
	local isChecked = self:GetChecked()
	TTB_SetTextboxBackground(isChecked)
end)

local IconModeCheckbox = TTB_CreateCheckbutton(ACPanel, 10, -150, "Icon mode")
IconModeCheckbox.tooltip = "When in combat, this will display a bar with the icons for your six buffs showing which are active. They are roughly ordered from best to worst (left to right)."

local IconBackgroundCheckbox = TTB_CreateCheckbutton(ACPanel, 180, -150, "Icon background")
IconBackgroundCheckbox.tooltip = "Show or hide the transparent black background in icon mode."

-- Preview the background of the icon box when checking the background box.
IconBackgroundCheckbox:SetScript("OnClick", function(self, ...)
	local isChecked = self:GetChecked()
	TTB_SetIconBackground(isChecked)
end)

local OrientationVCheckbox = TTB_CreateCheckbutton(ACPanel, 350, -150, "Vertical")
OrientationVCheckbox.tooltip = "Vertical icon alignment."

-- Pin checkboxes.
local TextboxPinCheckbox = TTB_CreateCheckbutton(ACPanel, 10, -180, "Pin text box")
TextboxPinCheckbox.tooltip = "Always show the text box."

local IconPinCheckbox = TTB_CreateCheckbutton(ACPanel, 180, -180, "Pin icons")
IconPinCheckbox.tooltip = "Always show the icons."

-- Lock checkboxes.
local TextboxLockCheckbox = TTB_CreateCheckbutton(ACPanel, 10, -210, "Lock text box")
TextboxLockCheckbox.tooltip = "Locking the text box gives you the ability to click through it."

local IconLockCheckbox = TTB_CreateCheckbutton(ACPanel, 180, -210, "Lock icons")
IconLockCheckbox.tooltip = "Locking the icons gives you the ability to click through them."

-- Threshold Slider.
local name = "Roll Threshold"
local template = "OptionsSliderTemplate"
local ACPThresholdSlider = CreateFrame("Slider", name, ACPanel, template)
ACPThresholdSlider:SetPoint("TOPLEFT", 10, -75)
ACPThresholdSlider.textLow = _G[name .. "Low"]
ACPThresholdSlider.textHigh = _G[name .. "High"]
ACPThresholdSlider.text = _G[name.."Text"]
ACPThresholdSlider:SetMinMaxValues(1, 25)
ACPThresholdSlider.minValue, ACPThresholdSlider.maxValue = ACPThresholdSlider:GetMinMaxValues() 
ACPThresholdSlider.textLow:SetText(ACPThresholdSlider.minValue)
ACPThresholdSlider.textHigh:SetText(ACPThresholdSlider.maxValue)
ACPThresholdSlider.text:SetText(name)
ACPThresholdSlider:EnableMouseWheel(true)
ACPThresholdSlider.tooltipText = "Sets the number of rolls you have to experience painfully before the addon starts yelling."
ACPThresholdSlider:SetValueStep(1)

-- Allow mouse wheel adjusting.
ACPThresholdSlider:SetScript("OnMouseWheel", function(self, delta)
	if delta > 0 then
		self:SetValue(self:GetValue() + self:GetValueStep())
	else
		self:SetValue(self:GetValue() - self:GetValueStep())
	end
end)

-- Update the slider description if its value changes.
ACPThresholdSlider:SetScript("OnValueChanged", function(self,event,arg1)
	ACPThresholdSlider.text:SetText("Roll Threshold (" .. math.floor(event) .. ")")
end)

-- Scale Slider.
local name = "Icon Scaling"
local template = "OptionsSliderTemplate"
local ACPScalingSlider = CreateFrame("Slider", name, ACPanel, template)
ACPScalingSlider:SetPoint("TOPLEFT", 200, -75)
ACPScalingSlider.textLow = _G[name .. "Low"]
ACPScalingSlider.textHigh = _G[name .. "High"]
ACPScalingSlider.text = _G[name.."Text"]
ACPScalingSlider:SetMinMaxValues(0.5, 2)
ACPScalingSlider.minValue, ACPScalingSlider.maxValue = ACPScalingSlider:GetMinMaxValues() 
ACPScalingSlider.textLow:SetText(ACPScalingSlider.minValue)
ACPScalingSlider.textHigh:SetText(ACPScalingSlider.maxValue)
ACPScalingSlider.text:SetText(name)
ACPScalingSlider:EnableMouseWheel(true)
ACPScalingSlider.tooltipText = "Scales the size of the icon mode."
ACPScalingSlider:SetValueStep(0.1)

-- Allow mouse wheel adjusting.
ACPScalingSlider:SetScript("OnMouseWheel", function(self, delta)
	if delta > 0 then
		self:SetValue(self:GetValue() + self:GetValueStep())
	else
		self:SetValue(self:GetValue() - self:GetValueStep())
	end
end)

-- Live preview of the icon bar when adjusting the scale slider.
ACPScalingSlider:SetScript("OnValueChanged", function(self, event, arg1)
	ACPScalingSlider.text:SetText("Icon Scaling (" .. TTB_Round(event, 1) .. ")")

	if(OrientationVCheckbox:GetChecked()) then
		orientation = "v"
	else
		orientation = "h"
	end

	-- Preview the icon box.
	TTB_DrawIcons(TTB_Round(event, 1), orientation)
end)

-- Preview the icon alignment. Has to know the ACPScalingSlider, so it's below it in the code.
OrientationVCheckbox:SetScript("OnClick", function(self, ...)
	local isChecked = self:GetChecked()

	if(isChecked) then
		TTB_DrawIcons(TTB_Round(ACPScalingSlider:GetValue(), 1), "v")
	else
		TTB_DrawIcons(TTB_Round(ACPScalingSlider:GetValue(), 1), "h")
	end
end)

-- Ham EditBox.
local ACPHamInputBox = CreateFrame("EditBox", nil, ACPanel, "InputBoxTemplate")
ACPHamInputBox:SetSize(250, 35)
ACPHamInputBox:SetPoint("TOPLEFT", 10, -260)
ACPHamInputBox:SetFontObject("GameFontHighlight")
ACPHamInputBox:SetTextInsets(0, 0, 0, 1)
ACPHamInputBox:SetAutoFocus(false)

local ACPHamInputBoxLabel = CreateFrame("Frame", "ACPHamInputBoxLabel", ACPanel)
ACPHamInputBoxLabel:SetSize(95, 20)
ACPHamInputBoxLabel:SetPoint("TOPLEFT", 10, -240)

ACPHamInputBoxLabel.text = ACPHamInputBoxLabel:CreateFontString(nil, "OVERLAY")
ACPHamInputBoxLabel.text:SetPoint("LEFT")
ACPHamInputBoxLabel.text:SetSize(95, 20)
ACPHamInputBoxLabel.text:SetFont("Fonts\\FRIZQT__.TTF", 12, "OUTLINE")
ACPHamInputBoxLabel.text:SetText("Ham message")

-- Finally EditBox.
local ACPFinallyInputBox = CreateFrame("EditBox", nil, ACPanel, "InputBoxTemplate")
ACPFinallyInputBox:SetSize(250, 35)
ACPFinallyInputBox:SetPoint("TOPLEFT", 10, -320)
ACPFinallyInputBox:SetFontObject("GameFontHighlight")
ACPFinallyInputBox:SetTextInsets(0, 0, 0, 1)
ACPFinallyInputBox:SetAutoFocus(false)

local ACPFinallyInputBoxLabel = CreateFrame("Frame", "ACPFinallyInputBoxLabel", ACPanel)
ACPFinallyInputBoxLabel:SetSize(105, 20)
ACPFinallyInputBoxLabel:SetPoint("TOPLEFT", 10, -300)

ACPFinallyInputBoxLabel.text = ACPFinallyInputBoxLabel:CreateFontString(nil, "OVERLAY")
ACPFinallyInputBoxLabel.text:SetPoint("LEFT")
ACPFinallyInputBoxLabel.text:SetSize(105, 20)
ACPFinallyInputBoxLabel.text:SetFont("Fonts\\FRIZQT__.TTF", 12, "OUTLINE")
ACPFinallyInputBoxLabel.text:SetText("Finally message")

-- Output channel dropdown menu.
local OutputChannelDropDownLabel = CreateFrame("Frame", "OutputChannelDropDownLabel", ACPanel)
OutputChannelDropDownLabel:SetSize(105, 20)
OutputChannelDropDownLabel:SetPoint("TOPLEFT", 10, -360)

OutputChannelDropDownLabel.text = OutputChannelDropDownLabel:CreateFontString(nil, "OVERLAY")
OutputChannelDropDownLabel.text:SetPoint("LEFT")
OutputChannelDropDownLabel.text:SetSize(105, 20)
OutputChannelDropDownLabel.text:SetFont("Fonts\\FRIZQT__.TTF", 12, "OUTLINE")
OutputChannelDropDownLabel.text:SetText("Output channel")

local OutputChannelDropDown = CreateFrame("Button", "OutputChannelDropDown", ACPanel, "UIDropDownMenuTemplate")
OutputChannelDropDown:SetPoint("TOPLEFT", 0, -380)
OutputChannelDropDown:Show()
 
local outputChannels = {
   "none",
   "say",
   "yell",
   "guild",
   "party",
   "raid",
}
 
function TTB_OnClick(self)
   UIDropDownMenu_SetSelectedID(OutputChannelDropDown, self:GetID())
   UIDropDownMenu_SetSelectedValue(OutputChannelDropDown, self.value)
end
 
function TTB_Initialize(self, level)
	local info = UIDropDownMenu_CreateInfo()

	for key, value in pairs(outputChannels) do
		info = UIDropDownMenu_CreateInfo()
		info.text = value
		info.value = value
		info.func = TTB_OnClick
		UIDropDownMenu_AddButton(info, level)
	end
end

UIDropDownMenu_Initialize(OutputChannelDropDown, TTB_Initialize)
UIDropDownMenu_SetWidth(OutputChannelDropDown, 120);
UIDropDownMenu_SetButtonWidth(OutputChannelDropDown, 120)
UIDropDownMenu_JustifyText(OutputChannelDropDown, "LEFT")

local iconModeBeforeOptionsMenu
local buffsOrderedBeforeOptionsMenu

--[[ This is called everytime when the options menu opens and sets its values.
	 When the options menu sometimes is blank, then most probably a wrong interface
	 element is addressed (e.g. PlaySoundCheck_B_ox instead of PlaySoundCheck_b_ox).
--]]
ACPanel.refresh = function()
	configMode = true

	--[[ Save the initial state of some saved variables. We want changes to them to be
		 visible during configuration and for the sake of simplicity do it via setting
		 their TTBDB values. If the user clicks 'Cancel', we set the TTBDB values
		 to these saved values again.
	--]]
	iconModeBeforeOptionsMenu = TTBDB.iconmode
	buffsOrderedBeforeOptionsMenu = TTBDB.BuffsOrdered

	LeeroyCheckbox:SetChecked(TTBDB.leeroy)
	PlaySoundCheckbox:SetChecked(TTBDB.playsound)

	TextboxModeCheckbox:SetChecked(TTBDB.textboxmode)
	TextboxBackgroundCheckbox:SetChecked(TTBDB.textboxbackground)

	IconModeCheckbox:SetChecked(TTBDB.iconmode)
	IconBackgroundCheckbox:SetChecked(TTBDB.iconbackground)

	TextboxPinCheckbox:SetChecked(TTBDB.pintextbox)
	IconPinCheckbox:SetChecked(TTBDB.pinicons)

	TextboxLockCheckbox:SetChecked(TTBDB.locktextbox)
	IconLockCheckbox:SetChecked(TTBDB.lockicons)

	ACPHamInputBox:SetText(TTBDB.hamMessage)
	ACPFinallyInputBox:SetText(TTBDB.finallyMessage)

	ACPHamInputBox:SetCursorPosition(0)
	ACPFinallyInputBox:SetCursorPosition(0)

	UIDropDownMenu_SetSelectedValue(OutputChannelDropDown, TTBDB.outputchannel, TTBDB.outputchannel)
	UIDropDownMenu_SetText(OutputChannelDropDown, TTBDB.outputchannel)

	if(TTBDB.icondir == "v") then
		OrientationVCheckbox:SetChecked(true)
	else
		OrientationVCheckbox:SetChecked(false)
	end
	
	--[[ The sliders have to know the status of the check boxes, so put them below.
		 Reason: When they get initialized by SetValue(), they fire their OnValueChanged event.
	--]]
	ACPThresholdSlider.text:SetText("Roll Threshold (" .. TTBDB.threshold .. ")")
	ACPThresholdSlider:SetValue(TTBDB.threshold)
	ACPScalingSlider.text:SetText("Icon Scaling (" .. TTBDB.iconscaling .. ")")
	ACPScalingSlider:SetValue(TTBDB.iconscaling)

	--[[ Needed to show the frames in other specs than Outlaw.
		 Showing them in Outlaw spec is done by setting the configMode above,
		 which affects TTB_OnUpdate, which is only active in Outlaw spec.
	--]]
	IconFrameContainer:Show()
	TextboxFrame:Show()

	-- Revert some preview changes made to elements after click on 'Cancel'.
	ACPanel.cancel = function(self)
		configMode = false

		TTBDB.BuffsOrdered = buffsOrderedBeforeOptionsMenu

		TTB_SetIconBackground(TTBDB.iconbackground)
		TTB_SetTextboxBackground(TTBDB.textboxbackground)

		TTB_DrawIcons(TTBDB.iconscaling, TTBDB.icondir)
		TTB_DrawTextbox()

		if(not TTBDB.pinicons or not isOutlaw) then
			IconFrameContainer:Hide()
		end

		if(not TTBDB.pintextbox or not isOutlaw) then
			TextboxFrame:Hide()
		end
	end
end

-- Save changes after click on 'Okay'.
ACPanel.okay = function(self)
	configMode = false

	TTBDB.leeroy = LeeroyCheckbox:GetChecked()
	TTBDB.playsound = PlaySoundCheckbox:GetChecked()
	TTBDB.textboxmode = TextboxModeCheckbox:GetChecked()

	if(TextboxModeCheckbox:GetChecked() == false) then
		TTBDB.pintextbox = false
		TextboxFrame:Hide()
	end

	TTBDB.iconmode = IconModeCheckbox:GetChecked()

	if(IconModeCheckbox:GetChecked() == false) then
		TTBDB.pinicons = false
		IconFrameContainer:Hide()
	end

	TTBDB.iconbackground = IconBackgroundCheckbox:GetChecked()
	TTB_SetIconBackground(TTBDB.iconbackground)

	TTBDB.textboxbackground = TextboxBackgroundCheckbox:GetChecked()
	TTB_SetTextboxBackground(TTBDB.textboxbackground)

	TTBDB.pintextbox = TextboxPinCheckbox:GetChecked()
	TTBDB.pinicons = IconPinCheckbox:GetChecked()

	if(TTBDB.pintextbox and isOutlaw) then
		TextboxFrame:Show()
	else
		TextboxFrame:Hide()
	end

	if(TTBDB.pinicons and isOutlaw) then
		IconFrameContainer:Show()
	else
		IconFrameContainer:Hide()
	end

	TTBDB.locktextbox = TextboxLockCheckbox:GetChecked()
	TTBDB.lockicons = IconLockCheckbox:GetChecked()

	TextboxFrame:EnableMouse(not TTBDB.locktextbox)
	IconFrameContainer:EnableMouse(not TTBDB.lockicons)

	if(OrientationVCheckbox:GetChecked()) then
		TTBDB.icondir = "v"
	else
		TTBDB.icondir = "h"
	end

	TTBDB.threshold = math.floor(ACPThresholdSlider:GetValue())
	TTBDB.iconscaling = TTB_Round(ACPScalingSlider:GetValue(), 1)
	TTBDB.hamMessage = ACPHamInputBox:GetText()
	TTBDB.finallyMessage = ACPFinallyInputBox:GetText()
	TTBDB.outputchannel = UIDropDownMenu_GetSelectedValue(OutputChannelDropDown)

	-- Refresh the GUI
	TTB_DrawIcons(TTBDB.iconscaling, TTBDB.icondir)
	TTB_DrawTextbox()

	--[[ Pretend as if we've just rolled when the icon mode is turned on.
		 Else no new cooldown animations would be placed on the icons.
	--]]
	if(iconModeBeforeOptionsMenu == false and TTBDB.iconmode) then
		JustRolled = true
	end

	-- Reset the frames when the corresponding mode was turned off.
	if(not TTBDB.iconmode) then
		TTB_ResetIconFrameContainer()
	end

	if(not TTBDB.textboxmode) then
		TTB_ResetTextboxFrame()
	end
end

--[[ Buff order editor. We draw a list of frames according to TTBDB.BuffsOrdered.
     We add every created frame to BuffOrderFrames. We add OnMouseUp handlers to them.
	 When we click one of those frames, it either gets pushed one index up (left click)
	 or one index down (right click) in the list.
--]]
local BuffOrderFrames = {}
function TTB_ACPInitBuffOrder()
	-- Create heading label.
	local BuffOrderHeadingLabel = CreateFrame("Frame", "BuffOrderHeadingLabel", ACPanel)
	BuffOrderHeadingLabel:SetSize(95, 20)
	BuffOrderHeadingLabel:SetPoint("TOPLEFT", 352, -240)

	BuffOrderHeadingLabel.text = BuffOrderHeadingLabel:CreateFontString(nil, "OVERLAY")
	BuffOrderHeadingLabel.text:SetPoint("LEFT")
	BuffOrderHeadingLabel.text:SetJustifyH("LEFT")
	BuffOrderHeadingLabel.text:SetSize(95, 20)
	BuffOrderHeadingLabel.text:SetFont("Fonts\\FRIZQT__.TTF", 12, "OUTLINE")
	BuffOrderHeadingLabel.text:SetText("Buff order:")
	BuffOrderHeadingLabel.text:SetTextColor(0.93, 0.96, 0.43, 1)

	-- Create footnode label.
	local BuffOrderFootnoteLabel = CreateFrame("Frame", "BuffOrderFootnoteLabel", ACPanel)
	BuffOrderFootnoteLabel:SetSize(200, 40)
	BuffOrderFootnoteLabel:SetPoint("TOPLEFT", 352, -420)

	BuffOrderFootnoteLabel.text = BuffOrderFootnoteLabel:CreateFontString(nil, "OVERLAY")
	BuffOrderFootnoteLabel.text:SetPoint("LEFT")
	BuffOrderFootnoteLabel.text:SetJustifyH("LEFT")
	BuffOrderFootnoteLabel.text:SetSize(200, 40)
	BuffOrderFootnoteLabel.text:SetFont("Fonts\\FRIZQT__.TTF", 10, "OUTLINE")
	BuffOrderFootnoteLabel.text:SetText("Note: Click left to move a buff up, right to move it down.")
	BuffOrderFootnoteLabel.text:SetTextColor(0.93, 0.96, 0.43, 1)

	i = 1
	offsetY = -268

	-- Create the list frames.
	for key, value in ipairs(TTBDB.BuffsOrdered) do
		BuffOrderFrames[i] = CreateFrame("Frame", "ACPBuff", ACPanel)
		BuffOrderFrames[i]:SetSize(150, 20)
		BuffOrderFrames[i]:SetPoint("TOPLEFT", 352, offsetY)

		BuffOrderFrames[i].text = BuffOrderFrames[i]:CreateFontString(nil, "OVERLAY")
		BuffOrderFrames[i].text:SetPoint("LEFT")
		BuffOrderFrames[i].text:SetJustifyH("LEFT")
		BuffOrderFrames[i].text:SetSize(150, 20)
		BuffOrderFrames[i].text:SetFont("Fonts\\FRIZQT__.TTF", 12, "OUTLINE")
		BuffOrderFrames[i].text:SetText(Buffs[value]["name"])

		BuffOrderFrames[i].index = i
		BuffOrderFrames[i].spellId = value

		-- We need to listen to OnMouseUp, OnClick is limited to buttons.
		BuffOrderFrames[i]:SetScript("OnMouseUp", function(self, button, ...)
			-- Push the buff one index up.
			if(button == "LeftButton") then
				-- We cannot push up the first item of the list.
				if(self.index == 1) then
					return
				end

				local thisText = BuffOrderFrames[self.index].text:GetText()
				local prevText = BuffOrderFrames[self.index - 1].text:GetText()

				local thisSpellId = BuffOrderFrames[self.index].spellId
				local prevSpellId = BuffOrderFrames[self.index - 1].spellId

				-- Switch values between the current frame and the one above
				BuffOrderFrames[self.index - 1].text:SetText(thisText)
				BuffOrderFrames[self.index].text:SetText(prevText)

				BuffOrderFrames[self.index - 1].spellId = thisSpellId
				BuffOrderFrames[self.index].spellId = prevSpellId

				TTB_RefreshOrderOfBuffs()
			-- Push the buff one index down.
			elseif(button == "RightButton") then
				-- We cannot push down the last item of the list.
				if(self.index == 6) then
					return
				end

				local thisText = BuffOrderFrames[self.index].text:GetText()
				local nextText = BuffOrderFrames[self.index + 1].text:GetText()

				local thisSpellId = BuffOrderFrames[self.index].spellId
				local nextSpellId = BuffOrderFrames[self.index + 1].spellId

				-- Switch values between the current frame and the one below.
				BuffOrderFrames[self.index + 1].text:SetText(thisText)
				BuffOrderFrames[self.index].text:SetText(nextText)

				BuffOrderFrames[self.index + 1].spellId = thisSpellId
				BuffOrderFrames[self.index].spellId = nextSpellId

				TTB_RefreshOrderOfBuffs()
			end
		end)

		-- Add hover effect OnEnter.
		BuffOrderFrames[i]:SetScript("OnEnter", function(self, button, ...)
			self.text:SetTextColor(0.7, 0.7, 0.7, 1)
		end)

		-- Remove hover effect OnLeave.
		BuffOrderFrames[i]:SetScript("OnLeave", function(self, button, ...)
			self.text:SetTextColor(1, 1, 1, 1)
		end)

		offsetY = offsetY - 25
		i = i + 1
	end
end

--[[ Updates the text box and icon bar to reflect changes made to the order of buffs.
     We can change the TTBDB.BuffsOrdered value directly, because the buff order can only be changed via
	 the GUI. If we click 'Cancel' in the GUI later, the value will be reset to the value before we
	 opened the GUI.
--]]
function TTB_RefreshOrderOfBuffs()
	TTBDB.BuffsOrdered = {}

	local i = 1
	for key, value in ipairs(BuffOrderFrames) do
		TTBDB.BuffsOrdered[i] = value.spellId
		i = i + 1
	end

	-- Get the current selected value from the GUI.
	local icondir
	if(OrientationVCheckbox:GetChecked()) then
		icondir = "v"
	else
		icondir = "h"
	end

	TTB_DrawIcons(TTB_Round(ACPScalingSlider:GetValue(), 1), icondir)
	TTB_DrawTextbox()
end

--[[ Centers the given frame horizontally.

	 @param Frame The frame to center
--]]
function TTB_CenterH(Frame)
	Frame:ClearAllPoints()
	Frame:SetPoint("BOTTOMLEFT", "UIParent", "BOTTOMLEFT", GetScreenWidth() / 2 - Frame:GetWidth() / 2, Frame:GetBottom())
end

--[[ Centers the given frame vertically.

	 @param Frame The frame to center
--]]
function TTB_CenterV(Frame)
	Frame:ClearAllPoints()
	Frame:SetPoint("BOTTOMLEFT", "UIParent", "BOTTOMLEFT", Frame:GetLeft(), GetScreenHeight() / 2 - Frame:GetHeight() / 2)
end

-- Manage centering on (ctrl) right clicking the icon bar.
IconFrameContainer:SetScript("OnMouseUp", function(self, button)
	if(button == "RightButton") then
		if(not IsControlKeyDown()) then
			TTB_CenterH(IconFrameContainer)
		else
			TTB_CenterV(IconFrameContainer)
		end
	end
end)

-- Manage centering on (ctrl) right clicking the text box.
TextboxFrame:SetScript("OnMouseUp", function(self, button)
	if(button == "RightButton") then
		if(not IsControlKeyDown()) then
			TTB_CenterH(TextboxFrame)
		else
			TTB_CenterV(TextboxFrame)
		end
	end
end)

--[[ Set standard saved variables as a table into TTBDB. Check if each entry exists
     and if not set standard values.
--]]
function TTB_SetStandardValues()
	if(TTBDB == nil) then
		TTBDB = {}
	end

	--[[ For retrieving the buffs in order (standard values are roughly best to worst).
	     These standard values should always represent the current meta.
	--]]
	if(TTBDB.BuffsOrdered == nil) then
		TTBDB.BuffsOrdered = {
			[1] = GrandMelee,
			[2] = RuthlessPrecision,
			[3] = TrueBearing,
			[4] = Broadsides,
			[5] = SkullAndCrossbones,
			[6] = BuriedTreasure,
		}
	end

	-- Message when rolling five buffs.
	if(TTBDB.hamMessage == nil) then
		TTBDB.hamMessage = "I rolled 'em all. Pulling Azeroth NOW!"
	end

	-- Message for ending a bad luck streak.
	if(TTBDB.finallyMessage == nil) then
		TTBDB.finallyMessage = "I was just trolled by the bones {n} times."
	end

	-- Output channel for messages.
	if(TTBDB.outputchannel == nil) then
		TTBDB.outputchannel = "yell"
	end

	-- Rolls since the last two buff roll.
	if(TTBDB.rolls == nil) then
		TTBDB.rolls = 0
	end

	-- The number of single buff rolls after the addon starts yelling when getting a two buff roll.
	if(TTBDB.threshold == nil) then
		TTBDB.threshold = 5
	end

	-- Scale factor of the icon bar.
	if(TTBDB.iconscaling == nil) then
		TTBDB.iconscaling = 1
	end

	-- Whether to play sounds that say the number of rolled buffs.
	if(TTBDB.playsound == nil) then
		TTBDB.playsound = false
	end

	-- Play 'Leeroy Jenkins' when getting five buffs.
	if(TTBDB.leeroy == nil) then
		TTBDB.leeroy = false
	end

	-- Record of not getting a two buff roll (single buff streak record).
	if(TTBDB.record == nil) then
		TTBDB.record = 0
	end

	-- Number of rolls done since the last five buff roll.
	if(TTBDB.lastfive == nil) then
		TTBDB.lastfive = 0
	end

	-- Timestamp of the last five roll.
	if(TTBDB.lastfivedate == nil) then
		TTBDB.lastfivedate = 0
	end

	-- Record number of rolls of not getting five rolls.
	if(TTBDB.lastfiverecord == nil) then
		TTBDB.lastfiverecord = 0
	end

	-- Text box mode.
	if(TTBDB.textboxmode == nil) then
		TTBDB.textboxmode = false
	end

	-- Pin state of the text box.
	if(TTBDB.pintextbox == nil) then
		TTBDB.pintextbox = false
	end

	-- Lock state of the text box.
	if(TTBDB.locktextbox == nil) then
		TTBDB.locktextbox = false
	end

	-- Background visibility of the text box.
	if(TTBDB.textboxbackground == nil) then
		TTBDB.textboxbackground = true
	end

	-- Icon mode.
	if(TTBDB.iconmode == nil) then
		TTBDB.iconmode = true
	end

	-- Pin state of the icon bar.
	if(TTBDB.pinicons == nil) then
		TTBDB.pinicons = false
	end

	-- Lock state of the icon bar.
	if(TTBDB.lockicons == nil) then
		TTBDB.lockicons = false
	end
	
	-- Background visibility of the icon bar.
	if(TTBDB.iconbackground == nil) then
		TTBDB.iconbackground = true
	end

	-- Direction of the icon bar.
	if(TTBDB.icondir == nil) then
		TTBDB.icondir = "h"
	end
end

--[[ This function rounds a number to the desired decimal place.

		@param number			The number to round
		@param digits			Decimal places
		
		@return Rounded number
--]]
function TTB_Round(number, digits)
	local shift = 10 ^ digits

	return floor(number * shift + 0.5) / shift
end

-- Slash commands.
SLASH_TTB1 = '/ttb'
function SlashCmdList.TTB(msg, editbox)
	local command, rest = msg:match("^(%S*)%s*(.-)$")

	-- No arguments, show the configuration GUI.
	if(command == "" and rest == "") then
		--[[ We need to call this twice, else only the interface menu will open
		     without navigating to our addon. For what reason ever.
		--]]
		InterfaceOptionsFrame_OpenToCategory(ACPanel.name)
		InterfaceOptionsFrame_OpenToCategory(ACPanel.name)
	-- Set the ham message.
	elseif(command == "ham" and rest ~= "") then
		TTBDB.hamMessage = rest
		ACPHamInputBox:SetText(TTBDB.hamMessage)
		print("Ham message set to: |cffEEF56E" .. rest .."|r")
	-- Print the ham message.
	elseif(command == "ham" and rest == "") then
		print("Ham message currently is: |cffEEF56E" .. TTBDB.hamMessage .. "|r")
	-- Set the finally message
	elseif(command == "finally" and rest ~= "") then 
		TTBDB.finallyMessage = rest
		print("Finally message set to: |cffEEF56E" .. rest .. "|r")
	-- Print the finally message.
	elseif(command == "finally" and rest == "") then
		print("Finally message currently is: |cffEEF56E" .. TTBDB.finallyMessage .. "|r")
	-- Set the threshold.
	elseif(command == "threshold" and rest ~= "" and tonumber(rest) ~= nil) then
		-- Check if its a valid threshold value.
		if(tonumber(rest) > 0 and tonumber(rest) <= 25) then
			TTBDB.threshold = tonumber(rest)

			-- Refresh the slider of the GUI configuration.
			ACPThresholdSlider:SetValue(TTBDB.threshold)

			print("Threshold set to: |cffEEF56E" .. rest .. "|r");
		else
			print("Please specify a number between |cffEEF56E1|r and |cffEEF56E25|r.")
		end
	-- Print the threshold.
	elseif(command == "threshold" and rest == "") then
		print("Threshold currently is: |cffEEF56E" .. TTBDB.threshold .. "|r")
	-- Set the icon scaling.
	elseif(command == "iconscaling" and rest ~= "" and tonumber(rest) ~= nil) then
		-- Check if its a valid scale factor.
		if(tonumber(rest) >= 0.5 and tonumber(rest) <= 2) then
			print("Icon scaling set to: |cffEEF56E" .. rest .. "|r");
			TTBDB.iconscaling = TTB_Round(tonumber(rest), 1)

			-- Refresh the slider of the GUI configuration and the icon bar.
			ACPScalingSlider:SetValue(TTBDB.iconscaling)
			TTB_DrawIcons(TTBDB.iconscaling, TTBDB.icondir)
		else
			print("Please specify a number between |cffEEF56E0.5|r and |cffEEF56E2|r.")
		end
	-- Print the scale factor.
	elseif(command == "iconscaling" and rest == "") then
		print("Icon scaling currently is: |cffEEF56E" .. TTBDB.iconscaling .. "|r")
	-- Set playsound.
	elseif(command == "playsound" and (rest == "false" or rest == "true")) then
		if(rest == "true") then
			TTBDB.playsound = true
			PlaySoundCheckbox:SetChecked(true)
		elseif(rest == "false") then
			TTBDB.playsound = false
			PlaySoundCheckbox:SetChecked(false)
		end

		print("Sound output set to: |cffEEF56E" .. tostring(TTBDB.playsound) .. "|r")
	-- Print playsound setting.
	elseif(command == "playsound" and rest == "") then
		print("Sound output currently is set to: |cffEEF56E" .. tostring(TTBDB.playsound) .. "|r")
	-- Print roll records.
	elseif(command == "record" and rest == "") then
		if(TTBDB.record > 0) then
			print("Your record of fishing for two buffs is |cffEEF56E" .. TTBDB.record .. " rolls|r.")
		end

		if(TTBDB.lastfiverecord > 0) then
			print("Your longest five roll dry spell was |cffEEF56E" .. TTBDB.lastfiverecord .. " rolls|r.")
		end

		if(TTBDB.lastfiverecord == 0 and TTBDB.record == 0) then
			print("You currently don't have any records. |TInterface\\Icons\\Ability_rogue_rollthebones:11:11:0:0:128:128:10:118:10:118|t |cffFF4500You better start rolling those bones!|r |TInterface\\Icons\\Ability_rogue_rollthebones:11:11:0:0:128:128:10:118:10:118|t")
		end
	-- Print how many rolls ago our last five roll happened.
	elseif(command == "lastfive" and rest == "") then
		local LastFiveDateText = ""

		if(TTBDB.lastfivedate > 0) then
			LastFiveDateText = " This was on |cffEEF56E" .. date("%m/%d/%y %H:%M:%S", TTBDB.lastfivedate) .. "|r."
		end

		if(TTBDB.lastfiverecord > 0) then
			print("The last time you've rolled 'em all was |cffEEF56E" .. TTBDB.lastfive + 1 .. " rolls|r ago." .. LastFiveDateText)
		else
			print("You'll need to roll them all at least once to show statistics.")
		end
	-- Set Leeroy mode.
	elseif(command == "leeroy" and (rest == "true" or rest == "false")) then
		if(rest == "true") then
			TTBDB.leeroy = true
			LeeroyCheckbox:SetChecked(true)
		elseif(rest == "false") then
			TTBDB.leeroy = false
			LeeroyCheckbox:SetChecked(false)
		end

		print("Leeroy mode set to: |cffEEF56E" .. tostring(TTBDB.leeroy) .. "|r")
	-- Print Leeroy mode setting.
	elseif(command == "leeroy" and rest == "") then
		print("Leeroy mode currently is set to: |cffEEF56E" .. tostring(TTBDB.leeroy) .."|r")
	-- Set text box mode.
	elseif(command == "textboxmode" and (rest == "true" or rest == "false")) then
		if(rest == "true") then
			TTBDB.textboxmode = true
			TextboxModeCheckbox:SetChecked(true)
		elseif(rest == "false") then
			TTBDB.textboxmode = false

			-- If the textbox mode gets disabled, we should set pintextbox to false.
			TTBDB.pintextbox = false
			TextboxModeCheckbox:SetChecked(false)
			TextboxFrame:Hide()
		end

		print("Text box mode set to: |cffEEF56E" .. tostring(TTBDB.textboxmode) .."|r")
	-- Set text box background visibility.
	elseif(command == "textboxbackground" and (rest == "true" or rest == "false")) then
		if(rest == "true") then
			TTBDB.textboxbackground = true
		elseif(rest == "false") then
			TTBDB.textboxbackground = false
		end

		TTB_SetTextboxBackground(TTBDB.textboxbackground)
		print("Text box background is set to: |cffEEF56E" .. tostring(TTBDB.textboxbackground) .. "|r")
	-- Print text box background setting.
	elseif(command == "textboxbackground" and rest == "") then
		print("Text box background currently is: |cffEEF56E" .. tostring(TTBDB.textboxbackground) .."|r")
	-- Lock the text box.
	elseif(command == "textbox" and rest == "lock") then
		TTBDB.locktextbox = true
		TextboxFrame:EnableMouse(false)
		print("The text box is now |cffEEF56Elocked|r. You can click through it.")
	-- Unlock the text box.
	elseif(command == "textbox" and rest == "unlock") then
		TTBDB.locktextbox = false
		TextboxFrame:EnableMouse(true)
		print("The text box is now |cffEEF56Eunlocked|r. You can move it around.")
	-- Pin the text box.
	elseif(command == "textbox" and rest == "pin") then
		TTBDB.pintextbox = true

		if(isOutlaw) then
			TextboxFrame:Show()
		end

		print("The text box is now |cffEEF56Epinned|r. It won't get hidden.")
	-- Unpin the text box.
	elseif(command == "textbox" and rest == "unpin") then
		TTBDB.pintextbox = false
		TextboxFrame:Hide()
		print("The text box is now |cffEEF56Eunpinned|r. It will get hidden after combat.")
	-- Set icon mode.
	elseif(command == "iconmode" and (rest == "true" or rest == "false")) then
		if(rest == "true") then
			TTBDB.iconmode = true
		elseif(rest == "false") then
			TTBDB.iconmode = false
			-- If the icon mode gets disabled, we should deactivate pinned icons
			TTBDB.pinicons = false
			IconFrameContainer:Hide()
		end

		print("Icon mode mode set to: |cffEEF56E" .. tostring(TTBDB.iconmode) .."|r")
	-- Print icon bar visibility.
	elseif(command == "iconbackground" and (rest == "true" or rest == "false")) then
		if(rest == "true") then
			TTBDB.iconbackground = true
		elseif(rest == "false") then
			TTBDB.iconbackground = false
		end

		TTB_SetIconBackground(TTBDB.iconbackground)
		print("Icon background is set to: |cffEEF56E" .. tostring(TTBDB.iconbackground) .."|r")
	-- Print icon background visibility.
	elseif(command == "iconbackground" and rest == "") then
		print("Icon background currently is: |cffEEF56E" .. tostring(TTBDB.iconbackground) .. "|r")
	-- Lock icon bar.
	elseif(command == "icons" and rest == "lock") then
		TTBDB.lockicons = true
		IconFrameContainer:EnableMouse(false)
		print("The icons are now |cffEEF56Elocked|r. You can click through them.")
	-- Unlock icon bar.
	elseif(command == "icons" and rest == "unlock") then
		TTBDB.lockicons = false
		IconFrameContainer:EnableMouse(true)
		print("The icons are now |cffEEF56Eunlocked|r. You can move them around.")
	-- Pin icon bar.
	elseif(command == "icons" and rest == "pin") then
		TTBDB.pinicons = true

		if(isOutlaw) then
			IconFrameContainer:Show()
		end

		print("The icons are now |cffEEF56Epinned|r. They won't get hidden.")
	-- Unpin icon bar.
	elseif(command == "icons" and rest == "unpin") then
		TTBDB.pinicons = false
		IconFrameContainer:Hide()
		print("The icons are now |cffEEF56Eunpinned|r. They will get hidden after combat.")
	-- Set icon direction.
	elseif(command == "icondir" and (rest == "h" or rest == "v")) then
		TTBDB.icondir = rest
		TTB_DrawIcons(TTBDB.iconscaling, TTBDB.icondir)
		print("Icon direction set to: |cffEEF56E" .. TTBDB.icondir .. "|r")
	-- Print icon direction
	elseif(command == "icondir" and rest == "") then
		print("Icon direction is currently set to: |cffEEF56E" .. TTBDB.icondir .. "|r")
	-- Reset frame positions.
	elseif(command == "resetpos") then
		IconFrameContainer:ClearAllPoints()
		IconFrameContainer:SetPoint("CENTER")
		TextboxFrame:ClearAllPoints()
		TextboxFrame:SetPoint("CENTER")
		print("The frames were |cffEEF56Ereset|r")
	-- Center frames horizontally.
	elseif(command == "center" and rest =="h") then
		TTB_CenterH(IconFrameContainer)
		TTB_CenterH(TextboxFrame)
		print("Centered |cffEEF56Ehorizontally|r.")
	-- Center frames vertically.
	elseif(command == "center" and rest == "v") then
		TTB_CenterV(IconFrameContainer)
		TTB_CenterV(TextboxFrame)
		print("Centered |cffEEF56Evertically|r.")
	-- Set the output channel.
	elseif(command == "output" and (rest == "none" or rest == "say" or rest == "yell"  or rest == "guild" or rest == "party" or rest == "raid")) then
		TTBDB.outputchannel = rest
		UIDropDownMenu_SetSelectedValue(OutputChannelDropDown, rest, rest)
		UIDropDownMenu_SetText(OutputChannelDropDown, rest)
		print("Output set to: |cffEEF56E" .. TTBDB.outputchannel .. "|r")
	-- Print output channel setting.
	elseif(command == "output" and (rest == "")) then
		print("Your current output channel is: |cffEEF56E" .. TTBDB.outputchannel .."|r")
	-- Reset some settings.
	elseif(command == "reset" and (rest == "")) then
		TTBDB.rolls = 0
		TTBDB.record = 0
		TTBDB.lastfive = 0
		TTBDB.lastfivedate = 0
		TTBDB.lastfiverecord = 0
		print("The statistics were |cffEEF56Ereset|r.")
	-- Print list of commands.
	elseif(command == "help" or command == "?") then
		DEFAULT_CHAT_FRAME:AddMessage("A list of |cffFF4500/ttb|r commands:", 1, 1, 1)
		DEFAULT_CHAT_FRAME:AddMessage("|cffFF4500Icon bar and textual output:|r", 1, 1, 1)
		DEFAULT_CHAT_FRAME:AddMessage("    textboxmode true/false |cffEEF56E- display buffs via text|r", 0.863, 0.863, 0.863)
		DEFAULT_CHAT_FRAME:AddMessage("    textbox pin/unpin |cffEEF56E- pin: always show the text box|r", 0.863, 0.863, 0.863)
		DEFAULT_CHAT_FRAME:AddMessage("    textbox lock/unlock |cffEEF56E- unlock: make the text box movable|r", 0.863, 0.863, 0.863)
		DEFAULT_CHAT_FRAME:AddMessage("    iconmode true/false |cffEEF56E- display buffs via icons|r", 0.863, 0.863, 0.863)
		DEFAULT_CHAT_FRAME:AddMessage("    icons pin/unpin |cffEEF56E- pin: always show the icons|r", 0.863, 0.863, 0.863)
		DEFAULT_CHAT_FRAME:AddMessage("    icons lock/unlock |cffEEF56E- unlock: make the icons movable|r", 0.863, 0.863, 0.863)
		DEFAULT_CHAT_FRAME:AddMessage("    icondir h/v |cffEEF56E- icon alignment: horizontal/vertical|r", 0.863, 0.863, 0.863)
		DEFAULT_CHAT_FRAME:AddMessage("|cffFF4500Positioning:|r", 1, 1, 1)
		DEFAULT_CHAT_FRAME:AddMessage("    center h/v |cffEEF56E- center frames horizontally/vertically|r", 0.863, 0.863, 0.863)
		DEFAULT_CHAT_FRAME:AddMessage("    resetpos |cffEEF56E- reset frame positions|r", 0.863, 0.863, 0.863)
		DEFAULT_CHAT_FRAME:AddMessage("|cffFF4500Yells and stuff:|r", 1, 1, 1)
		DEFAULT_CHAT_FRAME:AddMessage("    ham <message> |cffEEF56E- message when you've rolled 'em all|r", 0.863, 0.863, 0.863)
		DEFAULT_CHAT_FRAME:AddMessage("    finally <message> |cffEEF56E- message after ending a one roll streak|r", 0.863, 0.863, 0.863)
		DEFAULT_CHAT_FRAME:AddMessage("    output none/say/yell/guild/party/raid |cffEEF56E- output channel|r", 0.863, 0.863, 0.863)
		DEFAULT_CHAT_FRAME:AddMessage("|cffFF4500Misc:|r", 1, 1, 1)
		DEFAULT_CHAT_FRAME:AddMessage("    leeroy true/false |cffEEF56E- scream along!|r", 0.863, 0.863, 0.863)
		DEFAULT_CHAT_FRAME:AddMessage("    playsound true/false |cffEEF56E- play sounds when gaining buffs|r", 0.863, 0.863, 0.863)
		DEFAULT_CHAT_FRAME:AddMessage("    reset |cffEEF56E- reset statistics|r", 0.863, 0.863, 0.863)
		DEFAULT_CHAT_FRAME:AddMessage("For a full reference check: https://mods.curse.com/addons/wow/troll-the-bones.")
	else
		print("This is not a command of Troll The Bones. Please consider reading the documentation at https://mods.curse.com/addons/wow/troll-the-bones or type |cffFF4500/ttb help|r.")
	end
end